/*
 * h263codec.h
 *
 * H.323 protocol handler
 *
 * Open H323 Library
 *
 * Copyright (c) 1999-2000 Equivalence Pty. Ltd.
 *
  *
 */
 
/*
 * Initial release notes from Guilhem Tardy::
 *
 * Added support for video capabilities & codec, only tested under Linux!
 * The code for varying bit rate is copied from h261codec.cxx,
 * until it is moved to a separate file common to both video codecs.
 *
 */


#ifdef P_USE_PRAGMA
#pragma interface
#endif

#include <openh323buildopts.h>


#include "h323caps.h"

//#if defined(H323_VICH263) || defined(H323_RFC2190_AVCODEC)

class H264Decoder;
class H264Encoder;

///////////////////////////////////////////////////////////////////////////////

/**This class is a H.263 video capability.
 */

class H323_H264Capability : public H323VideoCapability
{
  PCLASSINFO(H323_H264Capability, H323VideoCapability)

  public:
  /**@name Construction */
  //@{
    /**Create a new H263 Capability
     */
    H323_H264Capability(
      unsigned sqcifMPI,
      unsigned qcifMPI,
      unsigned cifMPI,
      unsigned cif4MPI,
      unsigned cif16MPI,
      unsigned maxBitRate = 850,
      unsigned videoFrameRate = 25,
      PBoolean unrestrictedVector = FALSE,
      PBoolean arithmeticCoding = FALSE,
      PBoolean advancedPrediction = FALSE,
      PBoolean pbFrames = FALSE
    );
  //@}

  /**@name Overrides from class PObject */
  //@{
    /**Create a copy of the object.
      */
    virtual PObject * Clone() const;
  //@}

  
  /**@name Overrides from class PObject */
  //@{
    /**Compare object
      */
    Comparison Compare(const PObject & obj) const;
   //@}

  /**@name Identification functions */
  //@{
    /**Get the sub-type of the capability. This is a code dependent on the
       main type of the capability.

       This returns one of the four possible combinations of mode and speed
       using the enum values of the protocol ASN H245_AudioCapability class.
     */
    virtual unsigned GetSubType() const;

    /**Get the name of the media data format this class represents.
     */
    virtual PString GetFormatName() const;
  //@}

  /**@name Protocol manipulation */
  //@{
    /**This function is called whenever and outgoing TerminalCapabilitySet
       or OpenLogicalChannel PDU is being constructed for the control channel.
       It allows the capability to set the PDU fields from information in
       members specific to the class.

       The default behaviour sets the data rate field in the PDU.
     */
    virtual PBoolean OnSendingPDU(
      H245_VideoCapability & pdu  /// PDU to set information on
    ) const;

    /**This function is called whenever and outgoing RequestMode
       PDU is being constructed for the control channel. It allows the
       capability to set the PDU fields from information in members specific
       to the class.

       The default behaviour sets the resolution and bit rate.
     */
    virtual PBoolean OnSendingPDU(
      H245_VideoMode & pdu  /// PDU to set information on
    ) const;

    /**This function is called whenever and incoming TerminalCapabilitySet
       or OpenLogicalChannel PDU has been used to construct the control
       channel. It allows the capability to set from the PDU fields,
       information in members specific to the class.

       The default behaviour gets the data rate field from the PDU.
     */
    virtual PBoolean OnReceivedPDU(
      const H245_VideoCapability & pdu  /// PDU to set information on
    );

    /**Create the codec instance, allocating resources as required.
     */
    virtual H323Codec * CreateCodec(
      H323Codec::Direction direction  /// Direction in which this instance runs
    ) const;

#if 0
    /** Get sqcifMPI
     */
    unsigned GetSQCIFMPI() const
      { return sqcifMPI); }

    /** Get qcifMPI
     */
    unsigned GetQCIFMPI() const
      { return qcifMPI; }

    /** Get cifMPI
     */
    unsigned GetCIFMPI() const
      { return cifMPI; }

    /** Get cif4MPI
     */
    unsigned GetCIF4MPI() const
      { return cif4MPI; }

    /** Get cif16MPI
     */
    unsigned GetCIF16MPI() const
      { return cif16MPI; }

    /** Get maximum bit rate
     */
    unsigned GetMaxBitRate() const
      { return maxBitRate; }

    /** Get unrestrictedVector capabilty
     */
    PBoolean GetUnrestrictedVectorCapability() const
      { return unrestrictedVector; }

    /** Get arithmeticCoding capabilty
     */
    PBoolean GetArithmeticCodingCapability() const
      { return arithmeticCoding; }

    /** Get advancedPrediction capabilty
     */
    PBoolean GetAdvancedPredictionCapability() const
      { return advancedPrediction; }

    /** Get  pbFrames capabilty
     */
    PBoolean GetPbFramesCapability() const
      { return pbFrames; }

    /** Get temporal/spatial tradeoff capabilty
     */
    PBoolean GetTemporalSpatialTradeOffCapability() const
      { return temporalSpatialTradeOff; }

    /** Get hrd_B
     */
    PBoolean GetHrd_B() const
      { return hrd_B; }

    /** Get bppMaxKb
     */
    PBoolean GetBppMaxKb() const
      { return bppMaxKb; }

    /** Get slowSqcifMPI
     */
    unsigned GetSlowSQCIFMPI() const
      { return (sqcifMPI<0?-sqcifMPI:0); }

    /** Get slowQcifMPI
     */
    unsigned GetSlowQCIFMPI() const
      { return (qcifMPI<0?-qcifMPI:0); }

    /** Get slowCifMPI
     */
    unsigned GetSlowCIFMPI() const
      { return (cifMPI<0?-cifMPI:0); }

    /** Get slowCif4MPI
     */
    unsigned GetSlowCIF4MPI() const
      { return (cif4MPI<0?-cif4MPI:0); }

    /** Get slowCif16MPI
     */
    unsigned GetSlowCIF16MPI() const
      { return (cif16MPI<0?-cif16MPI:0); }

    /** Get errorCompensation capabilty
     */
    PBoolean GetErrorCompensationCapability() const
      { return errorCompensation; }
#endif

  //@}

protected:
    // H.263 v1
    signed sqcifMPI;		// {1..3600 units seconds/frame, 1..32 units 1/29.97 Hz}
    signed qcifMPI;		// {1..3600 units seconds/frame, 1..32 units 1/29.97 Hz}
    signed cifMPI;		// {1..3600 units seconds/frame, 1..32 units 1/29.97 Hz}
    signed cif4MPI;		// {1..3600 units seconds/frame, 1..32 units 1/29.97 Hz}
    signed cif16MPI;		// {1..3600 units seconds/frame, 1..32 units 1/29.97 Hz}

    unsigned maxBitRate;	// units of bit/s
    unsigned videoFrameRate;   // frames per second.

    PBoolean     unrestrictedVector;
    PBoolean     arithmeticCoding;
    PBoolean     advancedPrediction;
    PBoolean     pbFrames;
    PBoolean     temporalSpatialTradeOff;

    long unsigned hrd_B;	// units of 128 bits
    unsigned bppMaxKb;		// units of 1024 bits

    PBoolean     errorCompensation;
};

///////////////////////////////////////////////////////////////
/**This class is a H.263 video codec.
 */
class H323_H264Codec : public H323VideoCodec
{
  PCLASSINFO(H323_H264Codec, H323VideoCodec)

  public:
    /**Create a new H263 video codec
     */
    H323_H264Codec(
      Direction direction,	/// Direction in which this instance runs      
      unsigned sqcifMPI,
      unsigned qcifMPI,
      unsigned cifMPI,
      unsigned cif4MPI,
      unsigned cif16MPI,
      unsigned maxBitRate,
      unsigned videoFrameRate
    );

    ~H323_H264Codec();

    /**Encode the data from the appropriate device.
       This will encode a frame of data for transmission. The exact size and
       description of the data placed in the buffer is codec dependent but
       should be less than H323Capability::GetTxFramesInPacket() *
       OpalMediaFormat::GetFrameSize()  in length.

       The length parameter is filled with the actual length of the encoded
       data, often this will be the same as the size parameter.

       This function is called every GetFrameRate() timestamp units, so MUST
       take less than (or equal to) that amount of time to complete!

       Note that a returned length of zero indicates that time has passed but
       there is no data encoded. This is typically used for silence detection
       in an audio codec.
     */
    virtual PBoolean Read(
      BYTE * buffer,		/// Buffer of encoded data
      unsigned & length,	/// Actual length of encoded data buffer
      RTP_DataFrame & rtpFrame	/// RTP data frame
    );

    /**Decode the data and output it to appropriate device.
       This will decode a single frame of received data. The exact size and
       description of the data required in the buffer is codec dependent but
       should be less than H323Capability::GetRxFramesInPacket() *
       OpalMediaFormat::GetFrameSize()  in length.

       It is expected this function anunciates the data. That is, for example
       with audio data, the sound is output on a speaker.

       This function is called every GetFrameRate() timestamp units, so MUST
       take less than that amount of time to complete!
     */
    virtual PBoolean Write(
      const BYTE * buffer,	  /// Buffer of encoded data
      unsigned length,		  /// Length of encoded data buffer
      const RTP_DataFrame & rtp,  /// RTP data frame
      unsigned & written	  /// Number of bytes used from data buffer
    );

    /**Get the frame rate in RTP timestamp units.
      */
    virtual unsigned GetFrameRate() const { return timestampDelta; }

    /**Set the quality level of transmitted video data. 
       Is irrelevant when this codec is used to receive video data.
       Has a value of 1 (good quality) to 31 (poor quality).
       Quality is improved at the expense of bit rate.
    */
    void SetTxQualityLevel(int qLevel);
 
    /**Set the number of blocks in the background that need
       to be transmitted with each frame
    */
    void SetBackgroundFill(int fillLevel);

    /**Process a request for a new frame, 
       as part of the picture has been lost.
    */
    virtual void OnLostPartialPicture();

    /**
       Process a request for a new frame, 
       as all of the picture has been lost.
    */
    virtual void OnLostPicture();

    /**
       The ffmpeg library calls this routine, everytime it has a new packet to 
       send. Note that for every encode thread, this 1 method is called. Consequently, 
       code is required to separate out packets from different encode threads.
    */
    static void RtpCallback(void *data, int size, int packetNumber);

  protected:
    PBoolean Resize(int width, int height);

    PBoolean RenderFrame(); 
    PBoolean RenderFrame(const void * buffer);



    void InitialiseCodec();
    void CloseCodec();


    unsigned timestampDelta;

    H264Decoder *videoDecoder;
    H264Encoder *videoEncoder;

    int now;
    BYTE * rvts;
    int ndblk, nblk;
   
    PTime           startTime;
    PINDEX          bitsSent;

    unsigned        lastebits;
};

//#endif // H323_VICH263




/////////////////////////////////////////////////////////////////////////////
